/* 
**  Created by wplaat
**
**  Copyright (C) 2008 
**  ==================
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation, version 2.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <malloc.h>
#include <string.h>
#include <stdio.h>
#include <unistd.h>
#include <ogcsys.h>
#include <network.h>
#include <ogc/lwp_watchdog.h>
#include <sys/types.h>
#include <sys/errno.h>
#include <fcntl.h>

#include "http.h"

// -----------------------------------------------------------
// DEFINES
// -----------------------------------------------------------

#define DEBUG(X)            

#define TCP_CONNECT_TIMEOUT     5000
#define TCP_BLOCK_SIZE          (16 * 1024)
#define TCP_BLOCK_RECV_TIMEOUT  4000
#define TCP_BLOCK_SEND_TIMEOUT  4000
#define HTTP_TIMEOUT            300000
#define MAX_LEN        			100
#define NUM_THREADS             1

// -----------------------------------------------------------
// ENUMS
// -----------------------------------------------------------

enum
{
    TCP_INIT=0,
	TCP_CONNECTED1=1,
	TCP_CONNECTED2=2,
	TCP_ERROR=3,
	TCP_WAITING=4,
	TCP_END=5
};

typedef enum 
{
	HTTPR_OK,
	HTTPR_ERR_CONNECT,
	HTTPR_ERR_REQUEST,
	HTTPR_ERR_STATUS,
	HTTPR_ERR_TOOBIG,
	HTTPR_ERR_RECEIVE
} 
http_res;

// -----------------------------------------------------------
// VARIABLES
// -----------------------------------------------------------

char *http_host;
u16 http_port;
char *http_path;
u32 http_max_size;

http_res result;
u32 http_status;
u32 content_length;
u8 *http_data;

int retval;
u32 http_status;
u8  *outbuf;
u32 outlen;

int  tcp_state;
char version[MAX_LEN];
char tcp_token[MAX_LEN];
char tcp_url1[MAX_LEN];
char tcp_url2[MAX_LEN];

lwp_t threads[NUM_THREADS];
mutex_t mutexcheck;
mutex_t mutexversion;
bool do_tcp_treat;

// -----------------------------------------------------------
// METHODES
// -----------------------------------------------------------

s32 tcp_socket(void) 
{
        DEBUG( printf("tcp_socket:enter\n"); )
        s32 s;

        s = net_socket (AF_INET, SOCK_STREAM, IPPROTO_IP);
        if (s < 0) 
		{
			DEBUG ( printf("net_socket failed: %d\n",s); )
        }

	    DEBUG ( printf("tcp_socket: leave [%d]\n",s); )
        return s;
}

s32 tcp_connect(char *host, const u16 port)
{
	    DEBUG ( printf("tcp_connect:enter [host=%s|port=%d]\n",host,port); )
		
        struct hostent *hp;
        struct sockaddr_in sa;
        s32 s, res;
        s64 t;

        hp = net_gethostbyname(host);
        if (!hp || !(hp->h_addrtype == PF_INET)) 
		{            
			DEBUG ( printf("net_gethostbyname failed: %d\n",errno); )
            return errno;
        }

        s = tcp_socket();
        if (s < 0)
		{
		   DEBUG ( printf("Error creating socket, exiting"); )
           return s;
	    }

        memset (&sa, 0, sizeof (struct sockaddr_in));
        //sa.sin_family= PF_INET;
		sa.sin_family= AF_INET;
        sa.sin_len = sizeof (struct sockaddr_in);
        sa.sin_port= htons(port);
        memcpy ((char *) &sa.sin_addr, hp->h_addr_list[0], hp->h_length);

        DEBUG ( printf("DNS resolve=%d.%d.%d.%d\n", (unsigned char) hp->h_addr_list[0][0], (unsigned char) hp->h_addr_list[0][1], (unsigned char) hp->h_addr_list[0][2], (unsigned char)  hp->h_addr_list[0][3]); )

        t = gettime();
        while (true) 
		{
                if (ticks_to_millisecs (diff_ticks (t, gettime ())) > TCP_CONNECT_TIMEOUT) 
				{
					DEBUG ( printf("tcp_connect timeout\n"); )
                    net_close (s);
                    return -ETIMEDOUT;
                }

                res = net_connect (s, (struct sockaddr *) &sa, sizeof (struct sockaddr_in));

                if (res < 0) 
				{
                   if (res == -EISCONN)  break;

                   if (res == -EINPROGRESS || res == -EALREADY) 
				   {
                       usleep (20 * 1000);
                       continue;
                   }

   				   DEBUG ( printf("net_connect failed: %d\n",res); )
                   net_close(s);
                   return res;
                }
                break;
        }
		DEBUG ( printf("tcp_connect:leave\n"); )
        return s;
}

char * tcp_readln(const s32 s, const u16 max_length, const u64 start_time, const u16 timeout) 
{		
        char *buf;
        u16 c;
        s32 res;
        char *ret;

        buf = (char *) malloc (max_length);

        c = 0;
        ret = NULL;
        while (true) 
		{
                if (ticks_to_millisecs (diff_ticks (start_time, gettime ())) > timeout)
                        break;

                res = net_read (s, &buf[c], 1);

                if ((res == 0) || (res == -EAGAIN)) 
				{
                        usleep (20 * 1000);
                        continue;
                }

                if (res < 0) 
				{                       
						DEBUG ( printf("tcp_readln failed: %d\n", res); )
                        break;
                }

                if ((c > 0) && (buf[c - 1] == '\r') && (buf[c] == '\n')) 
				{
                        if (c == 1) 
						{
                                ret = strdup ("");
                                break;
                        }
                        ret = strndup (buf, c - 1);
                        break;
                }
                c++;
                if (c == max_length) break;
        }
        free (buf);
		return ret;
}

bool tcp_read(const s32 s, u8 **buffer, const u32 length) 
{
	    DEBUG ( printf("tcp_read:enter\n"); )
		
        u8 *p;
        u32 step, left, block, received;
        s64 t;
        s32 res;

        step = 0;
        p = *buffer;
        left = length;
        received = 0;

        t = gettime ();
        while (left) 
		{
                if (ticks_to_millisecs (diff_ticks (t, gettime ())) > TCP_BLOCK_RECV_TIMEOUT)
			    {
				   DEBUG ( printf("tcp_read timeout\n"); )
                   break;
                }

                block = left;
                if (block > 2048) block = 2048;
				
                res = net_read (s, p, block);
                if ((res == 0) || (res == -EAGAIN)) 
				{
                    usleep (20 * 1000);
                    continue;
                }

                if (res < 0) 
				{
					DEBUG ( printf("net_read failed %d\n", res); )
                    break;
                }

                received += res;
                left -= res;
                p += res;

                if ((received / TCP_BLOCK_SIZE) > step) 
				{
                    t = gettime ();
                    step++;
                }
        }

		DEBUG ( printf("tcp_read:leave\n"); )
        return left == 0;
}

bool tcp_write(const s32 s, const u8 *buffer, const u32 length)
{
	    DEBUG ( printf("tcp_read:write\n"); )
		
        const u8 *p;
        u32 step, left, block, sent;
        s64 t;
        s32 res;

        step = 0;
        p = buffer;
        left = length;
        sent = 0;

        t = gettime ();
        while (left) 
		{
                if (ticks_to_millisecs (diff_ticks (t, gettime ())) > TCP_BLOCK_SEND_TIMEOUT) 
				{
 					DEBUG ( printf("tcp_write timeout\n"); )
                    break;
                }

                block = left;
                if (block > 2048) block = 2048;

                res = net_write (s, p, block);
                if ((res == 0) || (res == -56))
				{
                    usleep (20 * 1000);
                    continue;
                }

                if (res < 0) 
				{
					DEBUG ( printf("net_write failed:%d\n",res); )
                    break;
                }
                sent += res;
                left -= res;
                p += res;

                if ((sent / TCP_BLOCK_SIZE) > step)
				{
                    t = gettime ();
                    step++;
                }
        }
        return left == 0;
		DEBUG ( printf("tcp_read:leave\n"); )
}

bool http_split_url(char **host, char **path, const char *url)
{
		DEBUG ( printf("http_split_url: enter\n"); )
		
        const char *p;
        char *c;

        if (strncasecmp (url, "http://", 7))
                return false;

        p = url + 7;
        c = strchr (p, '/');

        if (c[0] == 0)
                return false;

        *host = strndup (p, c - p);
        *path = strdup (c);

		DEBUG ( printf("http_split_url: leave\n"); )
		
        return true;
}

bool http_request(const char *url, const u32 max_size)
{
	    DEBUG ( printf("http_request: enter\n"); )
 		
        int linecount;
        if (!http_split_url(&http_host, &http_path, url)) return false;

        http_port = 80;
        http_max_size = max_size;
       
        http_status = 404;
        content_length = 0;
        http_data = NULL;

        int s = tcp_connect (http_host, http_port);

		DEBUG ( printf("tcp_connect(%s, %hu) = %d\n",http_host, http_port, s); )
        if (s < 0) 
		{
            result = HTTPR_ERR_CONNECT;
            return false;
        }

        char *request = (char *) malloc (1024);
        char *r = request;
        r += sprintf (r, "GET %s HTTP/1.1\r\n", http_path); 
        r += sprintf (r, "Host: %s\r\n", http_host);
		r += sprintf (r, "User-Agent: Wii/3.4 bibleQuiz/0.6\r\n");
        r += sprintf (r, "Accept: text/xml,text/html,text/plain\r\n");
        r += sprintf (r, "Cache-Control: no-cache\r\n\r\n");

        DEBUG ( printf("request = %s\n", request); )

        bool b = tcp_write (s, (u8 *) request, strlen (request));
		DEBUG ( printf("tcp_write returned=%d\n", b); )

        free (request);
        linecount = 0;

        for (linecount=0; linecount < 32; linecount++) 
		{
            char *line = tcp_readln ((s32) s, (u16) 0xff, (u64) gettime(), (u16) HTTP_TIMEOUT);
		    //DEBUG ( printf("tcp_readln: %s\n", line ? line:"(null)"); )
          
            if (!line) 
			{
                http_status = 404;
                result = HTTPR_ERR_REQUEST;
                break;
            }

            if (strlen (line) < 1) 
			{
                free (line);
                line = NULL;
                break;
            }

            sscanf (line, "HTTP/1.%*u %u", &http_status);
            sscanf (line, "Content-Length: %u", &content_length);

            free (line);
            line = NULL;
        }
        
		DEBUG ( printf("content_length = %d, status = %d, linecount = %d\n", content_length, http_status, linecount); )
      
        if (linecount == 32 || !content_length) http_status = 404;
        if (http_status != 200) 
		{
            result = HTTPR_ERR_STATUS;
            net_close (s);
            return false;
        }
        if (content_length > http_max_size) 
		{
            result = HTTPR_ERR_TOOBIG;
            net_close (s);
            return false;
        }
        http_data = (u8 *) malloc (content_length);
        b = tcp_read (s, &http_data, content_length);
        if (!b) 
		{
                free (http_data);
                http_data = NULL;
                result = HTTPR_ERR_RECEIVE;
                net_close (s);
                return false;
        }

        result = HTTPR_OK;
        net_close(s);

		DEBUG ( printf("http_request: leave\n"); )
        return true;
}

bool http_get_result(u32 *_http_status, u8 **content, u32 *length) 
{
		DEBUG ( printf("http_get_result: enter\n"); )
		
        if (http_status) *_http_status = http_status;

        if (result == HTTPR_OK) 
		{
           *content = http_data;
           *length = content_length;
        } 
		else 
		{
           *content = NULL;
           *length = 0;
        }

        free (http_host);
        free (http_path);

		DEBUG ( printf("http_get_result: leave\n"); )
        return true;
}

int tcp_init(void) 
{    
    DEBUG ( printf("tcp_init: enter\n"); )
	 	
    DEBUG ( printf("Waiting for network to initialise...\n"); )
			
	s32 result;
    while ((result = net_init()) == -EAGAIN);
	
    if ( result>= 0) 
    {
        char myIP[16];
        if (if_config(myIP, NULL, NULL, true) < 0) 
        {
            DEBUG ( printf("Error reading IP address, exiting"); )
			return 1;
        }
        DEBUG ( printf("Network initialised [%s].\n",myIP); )
    } 
    else 
    {
      DEBUG ( printf("Unable to initialise network, exiting"); )
	  return 1;
    }
	
	DEBUG ( printf("openNetwork: leave\n"); )
	return 0;
}


void tcp_sleep(unsigned int seconds)
{
   DEBUG ( printf("mssleep:enter\n"); )
   DEBUG ( printf("sleep %d sec.\n",seconds); )
   sleep(seconds);
   DEBUG ( printf("mssleep:leave\n"); )
}


char * findToken(u8 *buffer, int bufsize, char *token)
{
   DEBUG ( printf("findToken:enter [%s]\n",token); )
   
   int i, j=0;
   int startpos=-1;
   int endpos;
   int len=strlen(token);
   bool found=false;
   
   // find token begin
   startpos=-1;
   j=0;
   
   DEBUG ( printf("len=%d\n",len); )
	
   for (i=0; i<bufsize; i++)
   {
      if (buffer[i]==token[j])
	  {	     
	     if (startpos==-1) startpos=i;	
		 if (++j>=len) 
		 {  
		    DEBUG ( printf("startpos=%d\n",startpos); )
		    found=true;
			break;
	     }
	  }
	  else
	  {
	     // No token match, continue...
	     if (startpos!=-1)
		 {
		    j=0;
			startpos=-1;
	     }
	  }
   }
   
   if (found) 
   {
       startpos+=len;
	  
       // find end of token value   
       for (i=startpos; i<bufsize; i++)
       {	      
         if ((buffer[i]=='\n') || (buffer[i]=='\r') || (buffer[i]==' '))
	     {  	     
		    endpos=i;
			DEBUG ( printf("endpos=%d\n",endpos); )
			
			// (small hack) End string
			buffer[endpos]=0x00;
			
			DEBUG ( printf("findToken:leave [%s]\n", buffer+startpos); )
			return (char*) buffer+startpos;
	     }
	  }
   }   
   DEBUG ( printf("findToken:leave [NULL]\n"); )
   return NULL;
}


// -----------------------------------------------------------
// THREAD STUFF
// -----------------------------------------------------------
	
void *tcp_thread(void *threadid)
{
   DEBUG ( printf("tcp_thread:enter\n"); )

   int retval;
   u32 http_status;
   u8  *outbuf;
   u32 outlen;
  
   //int tid = (int)threadid;

   LWP_MutexLock (mutexcheck);
   while(do_tcp_treat)
   {
      LWP_MutexUnlock(mutexcheck);
	  
	  switch (tcp_state)
	  {
	    case TCP_INIT:   
		{
		  DEBUG ( printf("stateMachine=TCP_INIT\n"); )
		  retval=tcp_init();
		  if (retval==0) tcp_state=TCP_CONNECTED1; else tcp_state=TCP_ERROR;
		}
		break;
					 
		case TCP_CONNECTED1: 
		{		
		  DEBUG ( printf("stateMachine=TCP_CONNECTED1\n"); )
          retval = http_request(tcp_url1, 1 << 31);
          if (!retval) 
          {
	         DEBUG ( printf("Error making http request\n"); )
          }
          else
          {     
	         char *tmp;	
	         http_get_result(&http_status, &outbuf, &outlen);   	 			
			 tmp=findToken(outbuf, outlen, tcp_token);
			 DEBUG ( printf("version: %s\n", tmp ? tmp : "(null)" ); )
			 LWP_MutexLock (mutexversion);
			 if (tmp!=NULL) strncpy(version, tmp, sizeof(tmp));	 	 
			 LWP_MutexUnlock (mutexversion);
	         free(outbuf);			 
          } 	 
		  tcp_state=TCP_CONNECTED2;
		} 	
        break;	 
		
		case TCP_CONNECTED2: 
		{		
		  DEBUG ( printf("stateMachine=TCP_CONNECTED2\n"); )
          retval = http_request(tcp_url2, 1 << 31);
          if (!retval) 
          {
	         DEBUG ( printf("Error making http request2\n"); )
          }
          else
          {     
		     char *tmp;	
	         http_get_result(&http_status, &outbuf, &outlen);   	
			 tmp=findToken(outbuf, outlen, tcp_token);
			 DEBUG ( printf("version: %s\n", tmp ? tmp : "(null)" ); ) 
			 LWP_MutexLock (mutexversion);
			 if ((tmp!=NULL) && (version==NULL)) strncpy(version, tmp, sizeof(tmp));	 	
			 LWP_MutexUnlock (mutexversion);
			 free(outbuf);
          } 	 
		  tcp_state=TCP_END;
		} 	
        break;	 
		
	    case TCP_WAITING:
		{
		    DEBUG ( printf("stateMachine=TCP_WAITING\n"); )	        
            tcp_sleep(300);
			tcp_state=TCP_CONNECTED1;
	    }
		break;
		  
		case TCP_ERROR:
		{
		    DEBUG ( printf("stateMachine=TCP_ERROR\n"); )
	        DEBUG ( printf("sleep 60 sec.\n"); )
            tcp_sleep(60);
		    tcp_state=TCP_INIT;
	    }
		break;
		
		case TCP_END:
		{
		   DEBUG ( printf("stateMachine=TCP_END\n"); )
           do_tcp_treat=false;
		}		
        break;
		
      }
	  
      LWP_MutexLock(mutexcheck);
   }
   LWP_MutexUnlock(mutexcheck);
   //LWP_SuspendThread(tid);
   
   DEBUG ( printf("tcp_thread:leave [0]\n"); )
   return 0;
}

char * tcp_get_latest_available_version(void)
{
   LWP_MutexLock(mutexversion);
   return version;
   LWP_MutexUnlock(mutexversion);
}

int tcp_start_thread(char *url1, char *url2, char *token)
{
    DEBUG ( printf("tcp_start_thread:enter [%s]\n", url); )

    tcp_state=TCP_INIT;
    do_tcp_treat = true;

    memset(version,0x00,sizeof(version));
	strcpy(tcp_url1, url1);
	strcpy(tcp_url2, url2);
	strcpy(tcp_token, token);
	
	int rc = LWP_CreateThread(&threads[0], tcp_thread, NULL, NULL, 0, 1);
    if(rc!=0) 
	{
        DEBUG ( printf("ERROR; return code from pthread_create() is %d\n", rc); )
    }
	
	DEBUG ( printf("tcp_start_thread:leave [%d]\n",rc); )
	return rc;
}
	

int tcp_stop_thread(void)
{
   DEBUG ( printf("tcp_stop_thread:enter\n"); )

   LWP_MutexLock (mutexcheck);
   do_tcp_treat = false;
   LWP_MutexUnlock (mutexcheck);
   
   DEBUG ( printf("tcp_stop_thread:leave [0]\n"); )
   return 0;
}

// -----------------------------------------------------------
// THE END
// -----------------------------------------------------------

