/* 
**  RedSquare for Wii - Created by wplaat
**
**  RedSquare is an open-source 2D action game.
**
**  Copyright (C) 2008 
**  ==================
**
**  This program is free software; you can redistribute it and/or modify
**  it under the terms of the GNU General Public License as published by
**  the Free Software Foundation, version 2.
**
**  This program is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**  GNU General Public License for more details.
**
**  You should have received a copy of the GNU General Public License
**  along with this program; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
**
**  Release Notes:
**  ==============
**
**  29/11/2008 Version 1.0
**  - Start programming
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#include <math.h>
#include <ogcsys.h>
#include <gccore.h>
#include <gcmodplay.h> 
#include <wiiuse/wpad.h>
#include <jpeg/jpgogc.h>

#include <fcntl.h>
#include <unistd.h>
#include <time.h> 
 
#include "http.h"
#include "GRRLIB.h"

// -----------------------------------------------------------
// DEFINES
// -----------------------------------------------------------

#define PROGRAM_NAME	    "RedSquare"
#define PROGRAM_VERSION     "0.1"
#define RELEASE_DATE        "29-11-2008" 
#define URL1                "http://www.wiibrew.org/wiki/RedSquare"
#define URL2                "http://members.tele2.nl/wplaat/RedSquare.html"
#define URL_TOKEN           " Version "

#define WSP_POINTER_X       200
#define WSP_POINTER_Y       250

#define COLOR_LIGHTBLUE     0xEBCE87
#define COLOR_WHITESMOKE    0xFFFFFF
#define COLOR_DARKBLACK     0x000000

#define MAX_SQUARES   		5
#define MAX_HORZ_PIXELS		640
#define MAX_VERT_PIXELS		480

#define MIN_X_CURSOR		50
#define MIN_Y_CURSOR		50
#define MAX_X_CURSOR		590
#define MAX_Y_CURSOR		430

#define BUTTON_A            (WPAD_BUTTON_A     | WPAD_CLASSIC_BUTTON_A)
#define BUTTON_B            (WPAD_BUTTON_B     | WPAD_CLASSIC_BUTTON_B)
#define BUTTON_HOME         (WPAD_BUTTON_HOME  | WPAD_CLASSIC_BUTTON_HOME)
#define BUTTON_1            (WPAD_BUTTON_1     | WPAD_CLASSIC_BUTTON_X)
#define BUTTON_2            (WPAD_BUTTON_2     | WPAD_CLASSIC_BUTTON_Y)
#define BUTTON_MINUS        (WPAD_BUTTON_MINUS | WPAD_CLASSIC_BUTTON_MINUS)

// -----------------------------------------------------------
// PROTOTYPES
// -----------------------------------------------------------
 

// -----------------------------------------------------------
// TYPEDEFS and ENUMS
// -----------------------------------------------------------
 
typedef struct 
{
  // Png Image index
  u8 *redsquare;
  u8 *bluesquare1;
  u8 *bluesquare2;
  u8 *bluesquare3;
  u8 *bluesquare4;
  u8 *gameboard;
  u8 *pointer;
} 
image;

image images;

typedef struct
{
   u8 *image;
   int x;
   int y;
   int width;
   int height;
   bool directionX;
   bool directionY;
} 
square;

square squares[MAX_SQUARES];

// -----------------------------------------------------------
// VARIABLES
// -----------------------------------------------------------

u32     *frameBuffer[1] = {NULL};
static  GXRModeObj *rmode = NULL;
Mtx     GXmodelView2D;

int     yOffset;
int     yjpegOffset;
time_t  startTime;
int     speed;
bool    move;
time_t  playTime;
time_t  prevPlayTime;
char    tmp[256];
bool    selected;
 
// Intro1 Image
extern char     pic1data[];
extern int      pic1length;

// Intro2 Image
extern char     pic2data[];
extern int      pic2length;

// Intro3 Image
extern char     pic3data[];
extern int      pic3length;

// RedSquare1 Image
extern char     pic4data[];
extern int      pic4length;

// BlueSquare1 Image
extern char     pic5data[];
extern int      pic5length;

// BlueSquare2 Image
extern char     pic6data[];
extern int      pic6length;

// BlueSquare3 Image
extern char     pic7data[];
extern int      pic7length;

// BlueSquare4 Image
extern char     pic8data[];
extern int      pic8length;

// Pointer Image
extern char     pic9data[];
extern int      pic9length;

// Gameboard Image
extern char     pic10data[];
extern int      pic10length;

// -----------------------------------------------------------
// GAME LOGIC
// -----------------------------------------------------------

void moveSquares(void)
{ 
   int i;
   
   if (!move) return;
   
   for (i=1; i<MAX_SQUARES; i++)
   {
       if (squares[i].directionX)
	   {
           if (squares[i].x+squares[i].width<MAX_HORZ_PIXELS)
	       {
	          squares[i].x=squares[i].x+speed;	   
	       }
	       else
	       {
	          squares[i].directionX=false;
		      squares[i].x=squares[i].x-speed;	
	       }	
	   }
       else  	   
	   {
	       if (squares[i].x<=0)
	       {
	          squares[i].directionX=true;
		      squares[i].x=squares[i].x+speed;	
		   }
		   else
		   {
		      squares[i].x=squares[i].x-speed;	
		   }
	   }
	   
	   
	   if (squares[i].directionY)
	   {
           if (squares[i].y+squares[i].height<MAX_VERT_PIXELS)
	       {
	          squares[i].y=squares[i].y+speed;	   
	       }
	       else
	       {
	          squares[i].directionY=false;
		      squares[i].y=squares[i].y-speed;	
	       }	
	   }
       else  	   
	   {
	       if (squares[i].y<=0)
	       {
	          squares[i].directionY=true;
		      squares[i].y=squares[i].y+speed;	
		   }
		   else
		   {
		      squares[i].y=squares[i].y-speed;	
		   }
	   }
   }
}

bool collisionDetection(void)
{
     int i,x,y;
	  	 
	  // blue square collesion detection
	 for (i=1; i<MAX_SQUARES; i++)
	 { 		
	     for (x=squares[i].x; x<squares[i].x+squares[i].width; x++)
		 {
	        if (x>=squares[0].x) 
		    {
		       if (x<=squares[0].x+squares[0].width)			
               {			  
			      if (squares[i].y>=squares[0].y) 
			      {
				     if (squares[i].y<=squares[0].y+squares[0].height)
				     {
		               return true;		 
				     }
			      }
 
			      if (squares[i].y+squares[i].height>=squares[0].y) 
			      {
				     if (squares[i].y+squares[i].height<=squares[0].y+squares[0].height)
				     {
		                return true;		 
				     }
				  }
			   }
			}			
		 }
		 				 

         for (y=squares[i].y; y<squares[i].y+squares[i].height; y++)						 
		 {
   	        if (y>=squares[0].y) 
			{
			   if (y<=squares[0].y+squares[0].height)
			   {
			   	  if (squares[i].x>=squares[0].x) 
		          {
		              if (squares[i].x<=squares[0].x+squares[0].width)			
                      { 	
		                 return true;		 
					  }
				  }
				  if (squares[i].x+squares[i].width>=squares[0].x) 
		          {
		              if (squares[i].x+squares[i].width<=squares[0].x+squares[0].width)			
                      { 	
		                 return true;		 
					  }
				  }
			   }
			}
		 }		
	 }
	 	 
	 // Wall collesion detection
     if (squares[0].x<=MIN_X_CURSOR)
	 {
	    return true;     	 
	 }
	 
	 if (squares[0].x+squares[0].width+20>MAX_X_CURSOR)
	 {
	    return true;     	 
	 }
	 
	 if (squares[0].y<=MIN_Y_CURSOR)
	 {
	    return true;     	 
	 }
	 
     if (squares[0].y+squares[0].height>=MAX_Y_CURSOR)
	 {
	    return true;     	 
	 }
	 
	 return false;
}

void initThreads(void)
{ 
   tcp_start_thread(URL1, URL2, URL_TOKEN);
}

void initGame()
{
   startTime=time(NULL);
   speed=1;
   move=true;
   selected=false;
}

void initImages(void)
{
   // Buttons & Panels
   images.redsquare=GRRLIB_LoadTexture((unsigned char*) pic4data);
   images.bluesquare1=GRRLIB_LoadTexture((unsigned char*) pic5data);
   images.bluesquare2=GRRLIB_LoadTexture((unsigned char*) pic6data);
   images.bluesquare3=GRRLIB_LoadTexture((unsigned char*) pic7data);
   images.bluesquare4=GRRLIB_LoadTexture((unsigned char*) pic8data);
   images.pointer=GRRLIB_LoadTexture((unsigned char*) pic9data);
   images.gameboard=GRRLIB_LoadTexture((unsigned char*) pic10data);
}

void initSquares(void)
{
   // Red Square
   squares[0].image=images.redsquare;
   squares[0].width=64;
   squares[0].height=64;
   squares[0].x=(MAX_HORZ_PIXELS/2)-(squares[0].width/2);
   squares[0].y=(MAX_VERT_PIXELS/2)-(squares[0].height/2);
   
   // Blue Square 1
   squares[1].image=images.bluesquare1;
   squares[1].width=120;
   squares[1].height=64;
   squares[1].x=10;
   squares[1].y=10;
   squares[1].directionX=true;
   squares[1].directionY=true;
   
   // Blue Square 2
   squares[2].image=images.bluesquare2;
   squares[2].width=120;
   squares[2].height=32;
   squares[2].x=MAX_HORZ_PIXELS-squares[2].width-10;
   squares[2].y=10;
   squares[2].directionX=false;
   squares[2].directionY=true;
      
   // Blue Square 3
   squares[3].image=images.bluesquare3;
   squares[3].width=64;
   squares[3].height=120;
   squares[3].x=10;
   squares[3].y=MAX_VERT_PIXELS-squares[3].height-10;
   squares[3].directionX=false;
   squares[3].directionY=true;
      
   // Blue Square 4
   squares[4].image=images.bluesquare4;
   squares[4].width=64;
   squares[4].height=64;
   squares[4].x=MAX_HORZ_PIXELS-squares[4].width-10;
   squares[4].y=MAX_VERT_PIXELS-squares[4].height-10;   
   squares[4].directionX=false;
   squares[4].directionY=false;
}

// -----------------------------------------------------------
// BUTTON LOGIC
// -----------------------------------------------------------

void buttonA(int x, int y)
{
   if (( (x>=squares[0].x) && (x<=squares[0].x+squares[0].width) && 
	     (y>=squares[0].y) && (y<=squares[0].y+squares[0].height) )
	  || selected )
   {
   	 squares[0].x=x-40;
     squares[0].y=y-40;	
     selected=true;	  		   	   
   }
}
		
void buttonExit(int index)
{    
   // Stop network thread
   tcp_stop_thread();
	
   if (index==0)
   {   
	   exit(0);
   }
   else
   {
	   SYS_ResetSystem(SYS_RESTART,0,0);	
   }
}
 
// -----------------------------------------------------------
// SCREEN LOGIC
// -----------------------------------------------------------

void drawGameboard()
{ 	   
    int i=0;

    GRRLIB_DrawImg(0,0,MAX_HORZ_PIXELS, MAX_VERT_PIXELS, images.gameboard, 0, 1, 1, 255 );
      
	// draw all squares
	for (i=0; i<MAX_SQUARES; i++)
    {   
       GRRLIB_DrawImg(squares[i].x,squares[i].y,squares[i].width, squares[i].height, squares[i].image, 0, 1, 1, 255 );   
    }	
}

void drawJpegImage(char *pictureData, int pictureLength, int x, int y )
{
    JPEGIMG         jpeg;
    int             row,
                    col,
                    pix,
                    offset;
    unsigned int   *jpegout;

    memset(&jpeg, 0, sizeof(JPEGIMG));

    jpeg.inbuffer = pictureData;
    jpeg.inbufferlength = pictureLength;

    JPEG_Decompress(&jpeg);
	
    pix = 0;

    jpegout = (unsigned int *) jpeg.outbuffer;

    offset=(y+yjpegOffset)*320;
    for (row = 0; row < jpeg.height; row++)
    {
	    for (col = 0; col < (jpeg.width >> 1); col++)
		{
	       frameBuffer[0][offset + col + x] = jpegout[pix++];
		}
	    offset+=320;
    }

    free(jpeg.outbuffer);
}

void drawIntroScreen1(int channel)
{
   boolean flag = true;
   
   drawJpegImage(pic1data, pic1length, 0, 0 );
		   	
   // Handle buttons events
   while ( flag )
   {
		WPAD_ScanPads(); 
		if (WPAD_ButtonsDown(channel) & BUTTON_A) flag=false; 	
		VIDEO_SetNextFramebuffer(frameBuffer[0]); 
		VIDEO_Flush();
		VIDEO_WaitVSync();
   }
}

void drawIntroScreen2(int channel)
{
   boolean flag = true;
   
   drawJpegImage(pic2data, pic2length, 0, 0 );
		   	
   // Handle buttons events
   while ( flag )
   {
		WPAD_ScanPads(); 
		if (WPAD_ButtonsDown(channel) & BUTTON_A) flag=false; 	
		VIDEO_SetNextFramebuffer(frameBuffer[0]); 
		VIDEO_Flush();
		VIDEO_WaitVSync();
   }
}

void drawIntroScreen3(int channel)
{
   boolean flag = true;
   
   drawJpegImage(pic3data, pic3length, 0, 0 );
		   	
   // Handle buttons events
   while ( flag )
   {
		WPAD_ScanPads(); 
		if (WPAD_ButtonsDown(channel) & BUTTON_A) flag=false; 	
		VIDEO_SetNextFramebuffer(frameBuffer[0]); 
		VIDEO_Flush();
		VIDEO_WaitVSync();
   }
}

// -----------------------------------------------------------
// MAIN
// -----------------------------------------------------------

int main()
{
    int x=0, y=0;   
	
    VIDEO_Init();
	
    WPAD_Init();
    WPAD_SetIdleTimeout(60); // Wiimote is shutdown after 60 seconds of innactivity.
	WPAD_SetDataFormat(WPAD_CHAN_0, WPAD_FMT_BTNS_ACC_IR);	// enable accelerometers and IR
 
    // Obtain the preferred video mode from the system
	// This will correspond to the settings in the Wii menu
	rmode = VIDEO_GetPreferredMode(NULL);
	
	// allocate framebuffer (Needed for JPG welcome images)
	frameBuffer[0] = MEM_K0_TO_K1(SYS_AllocateFramebuffer(rmode));
		
	// Initialise the console, required for printf
	console_init(frameBuffer[0],20,20,rmode->fbWidth,rmode->xfbHeight,rmode->fbWidth*VI_DISPLAY_PIX_SZ);
	
	// Set up the video registers with the chosen mode
	VIDEO_Configure(rmode);
	
	if (rmode->xfbHeight==528)
	{
	    // TV mode PAL 50Hz
	    yOffset = 25;  
		yjpegOffset = 25;
	}
	else
	{      
	    // TV mode PAL 60Hz
	    yOffset = 25;
		yjpegOffset = 0;
	}

	// Tell the video hardware where our display memory is
	VIDEO_SetNextFramebuffer(frameBuffer[0]);
	
	// Make the display visible
	VIDEO_SetBlack(FALSE);

	// Flush the video register changes to the hardware
	VIDEO_Flush();

  	// Init game variables    
	initGame();
	initImages();
    initSquares();
	initThreads();
	GRRLIB_InitFreetype();
	
    // Show Intro screens
    drawIntroScreen1(0);	
	drawIntroScreen2(0);
	drawIntroScreen3(0);
	
    // Init GRRLib graphics library
    GRRLIB_InitVideo();
    GRRLIB_Start();
    
	// Make screen black
	GRRLIB_FillScreen(0xFFFFFF);
    GRRLIB_Render();
			
	// Repeat forever
    while(1)
	{					
        // Draw game board and other items
        drawGameboard();

        if (move)
		{  
		   playTime = time(NULL)-startTime;   
		   if (prevPlayTime!=playTime)
		   {
		      if ((playTime % 10)==0) 
			  {
			     speed++;			  
		         prevPlayTime=playTime;
			  }
		   }
		}
		   
        if (collisionDetection() || !move)
	    {	
   		   int xpos;	      
	       struct tm *level;
		   u8 *temp;

		   // Init text layer	  
           GRRLIB_initTexture();
		           
	       sprintf(tmp,"GAME OVER");
	       xpos=320-((strlen(tmp)*32)/2);	
	       GRRLIB_Printf2(xpos, 200, tmp, 48, COLOR_DARKBLACK);            

	       level = localtime(&playTime);
	       sprintf(tmp,"%02d:%02d:%02d", level->tm_hour, level->tm_min, level->tm_sec);
	       xpos=320-((strlen(tmp)*16)/2);	
	       GRRLIB_Printf2(xpos, 260, tmp, 24, COLOR_DARKBLACK);
			   	   
		   // Draw text layer on top of gameboard and squares
		   temp=GRRLIB_GetTexture();
           GRRLIB_DrawImg(0, 0, 640, 480, temp, 0, 1.0, 1.0, 255);	
		
           move=false;
	    }
	
	    moveSquares();
		
	    // Scan for button events
		WPAD_SetVRes(0, 640, 480);
        WPAD_ScanPads();
        u32 wpadheld = WPAD_ButtonsHeld(0);
		u32 wpaddown = WPAD_ButtonsDown(0);

        // Scan for ir events
        ir_t ir1; 
		WPAD_IR(WPAD_CHAN_0, &ir1); 
		x=ir1.sx-WSP_POINTER_X;
		y=ir1.sy-WSP_POINTER_Y;
					
     	// Draw wiiremote ir pointer
		GRRLIB_DrawImg( x, y, 96, 96, images.pointer, (ir1.angle/2), 1, 1, 255 );
		GRRLIB_Render();
		
		// handle button events
        if (wpadheld & BUTTON_A) 
		{  
		   buttonA(x+40,y+40);		
		}
		else selected=false;
	  
		if (wpaddown & BUTTON_B) 
		{
		   if (!move)
		   {
		      initSquares();
		      initGame();   
		   }
		}
		if (wpaddown & BUTTON_HOME) 
		{		   
		   buttonExit(0);
		}
		
	}
    return 0;
}

// -----------------------------------------------------------
// THE END
// -----------------------------------------------------------
